@@ -3,19 +3,19 @@ import numpy as np
 from .chart import plotly_heatmap, line_plotly
 from operator import attrgetter
 from .utils import filter_non_capitalized_words_from_list, structure_concated_dataframe
-import warnings
 from .clean import clean_class
 from .const import std_vals
 import seaborn as sns
 
 
 class eda_class:
-    def __init__(self, _df: pd.DataFrame, categorical_data: list = None) -> None:
+    def __init__(self, _df: pd.DataFrame) -> None:
         df = _df.copy()
         self.na_like_items_in_str = std_vals.na_like_items_in_str
 
         # Output dictionaries
         self.raw_input = df.copy()
+        self.SUMMARY = {}
         self.DTYPES = {}
         self.DOWNCASTING = {}
         self.MISSING_VALUES = {}
@@ -24,6 +24,10 @@ class eda_class:
 
         # OUTPUT
 
+        # Summary
+        self.SUMMARY["info"] = self.generate_df_info()
+        self.SUMMARY["describe"] = self.generate_dataframe_describe(_df=self.raw_input)
+
         # Identifying dtypes
         self.DTYPES["boolean_like_columns"] = self.identify_boolean_like_columns()
         self.DTYPES["datetime"] = self.identify_datetime_columns()
@@ -33,9 +37,7 @@ class eda_class:
         self.DTYPES["potential_uid_columns"] = self.find_unique_identifier_columns()
 
         # Interim calculations
-        self.correlation_matrix_numerical = self.make_upper_triangle_na(
-            _df=self.make_diagonals_na(_df=self.raw_input.corr())
-        )
+        self.correlation_matrix_numerical = self.raw_input.corr()
         self.correlation_matrix_non_numerical = (
             self.get_correlation_of_non_numerical_columns()
         )
@@ -47,31 +49,27 @@ class eda_class:
         ] = self.determine_candidates_for_float32_conversion()
 
         # Missing values
-        self.MISSING_VALUES[
-            "columns_with_no_null_values"
-        ] = self.generate_list_of_columns_with_no_null_values()
-        self.MISSING_VALUES[
-            "columns_with_null_values"
-        ] = self.generate_list_of_columns_with_null_values()
-        self.MISSING_VALUES[
-            "percentage_of_missing_values"
-        ] = self.generate_dataframe_with_percentage_of_missing_values()
-        self.MISSING_VALUES[
-            "percentage_of_non_missing_values"
-        ] = self.generate_dataframe_with_percentage_of_non_missing_values()
         self.MISSING_VALUES[
             "dates_continuity_check"
         ] = self.identify_if_dates_are_continuous()
         self.MISSING_VALUES["plotly_missing_values_heatmap"] = plotly_heatmap(
-            _df=self.raw_input.isnull(), title="Missing Values Heatmap"
+            _df=self.raw_input.isnull(),
+            title="Missing Values Heatmap",
+            show_y_axis=False,
+            show_x_axis=True,
         )
         self.MISSING_VALUES["plotly_correlation_missing_values"] = plotly_heatmap(
             _df=self.generate_correlation_of_missing_values(),
             title="Correlation of Missing Values",
+            show_y_axis=True,
+            show_x_axis=False,
         )
         self.MISSING_VALUES[
             "na_like_values_in_str_columns"
         ] = self.get_mask_for_na_like_items_in_str_columns()
+        self.MISSING_VALUES[
+            "na_in_datetime_columns"
+        ] = self.identify_na_in_datetime_column()
 
         # Duplicates
         self.DUPLICATES["redundant_columns"] = self.identify_redundant_columns()
@@ -81,27 +79,39 @@ class eda_class:
         self.DATA_ANALYSIS["plotly_correlation_numerical"] = plotly_heatmap(
             _df=self.correlation_matrix_numerical,
             title="Correlation Matrix Numerical",
+            show_y_axis=True,
+            show_x_axis=False,
         )
+
         self.DATA_ANALYSIS["plotly_correlation_non_numerical"] = plotly_heatmap(
             _df=self.correlation_matrix_non_numerical,
             title="Correlation Matrix Non Numerical",
+            show_y_axis=True,
+            show_x_axis=False,
         )
+
         self.DATA_ANALYSIS["histogram_matplotlib"] = self.histograms_numeric_columns(
             numerical_columns=self.raw_input.select_dtypes(include="number"),
         ).figure
 
-        # Categorical plot
-        if categorical_data is None:
-            self.categorical_data = self.DTYPES["categorical_like_columns"]
-        else:
-            self.categorical_data = categorical_data
-        self.DATA_ANALYSIS["describe"] = self.generate_dataframe_describe(
-            _df=self.raw_input
-        )
         self.DATA_ANALYSIS[
             "top_10_most_frequent_values"
         ] = self.identify_top_n_most_frequent_value_across_non_numeric_columns(n=10)
 
+    def generate_df_info(self) -> pd.DataFrame:
+        df = self.raw_input.copy()
+        df_info = pd.DataFrame(
+            {
+                "columns": df.columns,
+                "non_nulls": len(df) - df.isnull().sum().values,
+                "nulls": df.isnull().sum().values,
+                "type": df.dtypes.values,
+            }
+        )
+        df_info["non_nulls_%"] = (df_info["non_nulls"] / len(df)).round(2)
+        df_info["nulls_%"] = (df_info["nulls"] / len(df)).round(2)
+        return df_info
+
     # Reference: eda_and_beyond
     def histograms_numeric_columns(self, numerical_columns: list):
         """
@@ -114,6 +124,18 @@ class eda_class:
         g = g.map(sns.distplot, "value")
         return g
 
+    def identify_na_in_datetime_column(self) -> dict:
+        df = self.raw_input.copy()
+        datetime_cols = self.DTYPES["datetime"]
+        df = df[datetime_cols]
+        na_in_datetime_cols_dict = {}
+        for col in datetime_cols:
+            df[col] = pd.to_datetime(df[col], errors="coerce")
+            mask = df[col].isna()
+            df_col = df[mask]
+            na_in_datetime_cols_dict[col] = df_col.index.to_list()
+        return na_in_datetime_cols_dict
+
     def get_mask_for_na_like_items_in_str_columns(self) -> pd.DataFrame:
         # Select the requisite columns
         df = self.raw_input.copy()
@@ -249,43 +271,11 @@ class eda_class:
 
         return df_describe
 
-    def generate_list_of_columns_with_null_values(self) -> list:
-        df = self.raw_input.copy()
-        return df.columns[df.isnull().any()].tolist()
-
-    def generate_list_of_columns_with_no_null_values(self) -> list:
-        df = self.raw_input.copy()
-        return df.columns[df.notnull().all()].tolist()
-
     def generate_correlation_of_missing_values(self) -> pd.DataFrame:
         df_null = self.raw_input.copy().isnull()
         df_null = df_null[df_null.any(axis=1)]
         return df_null.corr()
 
-    def generate_dataframe_with_percentage_of_missing_values(
-        self, without_formating: bool = False
-    ) -> pd.DataFrame:
-        df = self.raw_input.copy()
-        df_output = pd.DataFrame(
-            df.isnull().sum() / len(df), columns=["percentage_missing_values"]
-        )
-        if without_formating:
-            return df_output
-        df_output = self.convert_float_to_percentage_in_dataframe(
-            _df=df_output, cols=["percentage_missing_values"]
-        )
-        return structure_concated_dataframe(df_output)
-
-    def generate_dataframe_with_percentage_of_non_missing_values(self) -> pd.DataFrame:
-        df_output = self.generate_dataframe_with_percentage_of_missing_values(
-            without_formating=True
-        )
-        df_output = df_output.rename(
-            columns={"percentage_missing_values": "percentage_non_missing_values"}
-        )
-        df_output = 1 - df_output
-        return structure_concated_dataframe(df_output)
-
     def convert_float_to_percentage_in_dataframe(
         self, _df: pd.DataFrame, cols: list
     ) -> pd.DataFrame:
@@ -358,12 +348,6 @@ class eda_class:
             .index.to_list()
         )
 
-    def make_upper_triangle_na(self, _df: pd.DataFrame) -> pd.DataFrame:
-        df = _df.copy()
-        mask = np.triu(np.ones_like(df, dtype=bool))
-        df = df.mask(mask)
-        return df
-
     def identify_categorical_like_columns(self) -> list:
         df = self.raw_input.copy()
         categorical_like_columns = []
@@ -382,18 +366,11 @@ class eda_class:
         )
         return categorical_like_columns
 
-    def make_diagonals_na(self, _df: pd.DataFrame) -> pd.DataFrame:
-        df = _df.copy()
-        # fluctuating diagonals
-        df.values[[np.arange(df.shape[0])] * 2] = np.nan
-        return df
-
     def get_correlation_of_non_numerical_columns(self) -> pd.DataFrame:
         df = self.raw_input.copy()
         df = df.select_dtypes(exclude=[np.number])
         df = df.apply(lambda col: pd.factorize(col)[0])
-        df_corr = self.make_diagonals_na(_df=df.corr())
-        df_corr = self.make_upper_triangle_na(_df=df_corr)
+        df_corr = df.corr()
         return df_corr
 
     def identify_datetime_columns(self) -> pd.DataFrame:
@@ -414,6 +391,18 @@ class eda_class:
         all_attributes = self.get_attributes_of_class()
         return self.filter_non_capitalized_words_from_list(_list=all_attributes)
 
+    def get_methods_of_class(self) -> list:
+        return [
+            attr
+            for attr in dir(self)
+            if callable(getattr(self, attr)) and not attr.startswith("__")
+        ]
+
+    def get_additional_eda_functionality(self) -> list:
+        all_attributes = self.get_methods_of_class()
+        # Identify items ending with "_"
+        return [attr for attr in all_attributes if attr.endswith("_")]
+
     def determine_candidates_for_float32_conversion(self) -> list:
         df = self.raw_input.copy()
         float32_cols = []
@@ -479,7 +468,7 @@ class eda_class:
 
         return df
 
-    def cdf_plotly(self, _df: pd.DataFrame, col: str):
+    def cdf_plotly_(self, _df: pd.DataFrame, col: str):
         df_ecdf = self.calculate_empirical_cdf(_df=_df, col=col)
 
         # generate_chart
@@ -493,7 +482,7 @@ class eda_class:
 
         return fig
 
-    def generate_describe_after_groupby(
+    def generate_describe_after_groupby_(
         self, _df: pd.DataFrame, groupby_col_name: list, describe_col_name: str
     ):
         """
